/*************************************************************************
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 2007      Bryan Christ <bryan.christ@hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <locale.h>
#include <inttypes.h>

#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>

#ifdef __linux
#include <sys/klog.h>
#endif

#include <glib.h>
#include <gmodule.h>

#ifdef _VIPER_WIDE
#include <ncursesw/curses.h>
#else
#include <curses.h>
#endif

#include <menu.h>
#include <pseudo.h>
#include <viper.h>

#include "vwm.h"
#include "vwm_private.h"
#include "vwm_menu.h"
#include "vwm_panel.h"
#include "vwm_scrsaver.h"
#include "vwm_modules.h"
#include "vwm_psthread.h"
#include "vwm_settings.h"
#include "vwm_signals.h"

ps_runq_t		*vwm_runq=NULL;
sig_atomic_t	vwm_task_count=0;	/* in practice same as int	*/

int main(void)
{
   extern WINDOW  		*SCREEN_WINDOW;
	extern ps_runq_t		*vwm_runq;
	WINDOW	      		*screen_window;
	WINDOW	      		*bg;
	struct timespec  		sleep_value={.tv_sec=0,.tv_nsec=1000};
	gint						ui_speed;
	gchar		      		*msg;
	gint		      		fd;
	gchar		      		*locale=NULL;
   uintmax_t      		msg_id=0;
	int						flags;

   /* abort when dlopen() is not supported  */
   if(g_module_supported()==FALSE)
   {
      printf("%s\n\r%s",
         "FATAL ERROR:  This system does not support dynamic module loading!",
         "Aborting.");
      exit(EXIT_FAILURE);
   }

	/*	set the locale to the default settings (as configured by env).
		this is required for ncurses to work properly.	*/	
	locale=setlocale(LC_ALL,"");
	/*	print some debug information.	*/
	printf("%s\n\r",locale);
	printf("ncurses = %d.%d (%d)\n\r",NCURSES_VERSION_MAJOR,
		NCURSES_VERSION_MINOR,NCURSES_VERSION_PATCH);
	fflush(NULL);

#ifdef __linux
   /*  suppress printk messages.  klogctl() is linux specific. */
	klogctl(6,NULL,0);
   printf("VWM running on Linux\n\r");
#endif

   /* supress STDERR */
	fd=open("/dev/null",O_WRONLY);
	if(fd==-1) exit(0);
	dup2(fd,STDERR_FILENO);

	/* ignore terminal interrupt signal */
   vwm_sigset(SIGINT,SIG_IGN);

#ifdef _DEBUG
   vwm_sigset(SIGILL,vwm_backtrace);
   vwm_sigset(SIGSEGV,vwm_backtrace);
   vwm_sigset(SIGFPE,vwm_backtrace);
#endif

	vwm_sigset(SIGIO,vwm_SIGIO);
	fcntl(STDIN_FILENO,F_SETOWN,getpid());
	flags=fcntl(STDIN_FILENO,F_GETFL);
	fcntl(STDIN_FILENO,F_SETFL, flags | FASYNC);

	/* start thread system */
	g_thread_init(NULL);

   viper_init(VIPER_GPM_SIGIO);
	screen_window=SCREEN_WINDOW;
   viper_set_border_agent(vwm_default_border_agent_unfocus,0);
   viper_set_border_agent(vwm_default_border_agent_focus,1);

	/* use the integrated window manager */
	vwm_init();
   vwm_panel_init();
	vwm_hook_enter(vwm_hook_wm_start,(gpointer)&msg_id);
	vwm_hook_leave(vwm_hook_wm_stop,(gpointer)&msg_id);

	/*	recycle the existing bacground window created by libviper at
		init.	we could create our own window, but it's not necessary.	*/
	bg=viper_screen_get_wallpaper();
	vwm_background(bg,(gpointer)0);
	viper_screen_set_wallpaper(bg,vwm_background,(gpointer)0);
	viper_screen_redraw(REDRAW_ALL);

   vwm_modules_preload();

	vwm_runq=psthread_init(PS_SCHED_LEAPFROG);
	psthread_add(vwm_runq,vwm_clock_driver,NULL);
	psthread_add(vwm_runq,vwm_poll_input,NULL);

/* this will load the default screensaver but it will be immediately
   overridden if by vwm_settings_load() if the user has specified something
   different in their ~/.vwm/vwmrc file.  */
#ifdef _VWM_SCREENSAVER_TIMEOUT
   vwm_scrsaver_timeout_set(_VWM_SCREENSAVER_TIMEOUT);
   vwm_scrsaver_set("SysSaver");
#endif

   vwm_settings_load(VWM_RC_FILE);
   vwm_scrsaver_start();

   vwm_panel_message_add(VWM_MAIN_MENU_HELP,-1);

	while(TRUE)
	{
		msg=(gchar*)shmq_msg_get("ui");
		if(msg!=NULL) if(strcmp(msg,"shutdown vwm")==0) break;

		ui_speed=vwm_ui_get_speed();
		if(ui_speed>1)
		{
			psthread_run(vwm_runq,ui_speed);
			vwm_ui_set_speed(1);
		}
		else psthread_run(vwm_runq,1);

		nanosleep(&sleep_value,NULL);
	}

	curs_set(1);
	endwin();
	close(fd);
	return 0;
}

VWM* vwm_init(void)
{
	static VWM	*vwm=NULL;

	if(vwm==NULL)
	{
 		vwm=(VWM*)g_malloc0(sizeof(VWM));

		vwm_keyring_set(1);
		/*	set control-w to start window manager when keyring 1 is active	*/
		vwm_hotkey_set(VWM_HOTKEY_WM,vwm_default_key_VWM_START,1);
		/*	set control-w to exit window manager when keyring 0 is active	*/
		vwm_hotkey_set(VWM_HOTKEY_WM,vwm_default_key_VWM_STOP,0);

		vwm_hotkey_set(VWM_HOTKEY_MENU,vwm_main_menu_hotkey,1);

		vwm_hotkey_set(17,vwm_default_key_WINDOW_CLOSE,0);
		vwm_hotkey_set(KEY_TAB,vwm_default_key_WINDOW_CYCLE,0);
		vwm_hotkey_set(KEY_UP,vwm_default_key_WINDOW_MOVE_UP,0);
		vwm_hotkey_set(KEY_DOWN,vwm_default_key_WINDOW_MOVE_DOWN,0);
		vwm_hotkey_set(KEY_LEFT,vwm_default_key_WINDOW_MOVE_LEFT,0);
		vwm_hotkey_set(KEY_RIGHT,vwm_default_key_WINDOW_MOVE_RIGHT,0);
	}

	return vwm;
}

void vwm_ui_accel(gint val)
{
	extern sig_atomic_t	vwm_task_count;

	if(val<1) val=1;

	vwm_task_count+=val;
}

void vwm_ui_slow(gint val)
{
	extern sig_atomic_t  vwm_task_count;

   if(val<1) val=1;

   vwm_task_count-=val;
	if(vwm_task_count<0) vwm_task_count=0;
}

gint vwm_ui_get_speed(void)
{
	extern sig_atomic_t vwm_task_count;

	return (gint)vwm_task_count;
}

void vwm_ui_set_speed(guint speed)
{
	extern sig_atomic_t vwm_task_count;

	vwm_task_count=speed;
}

void vwm_hook_enter(VIPER_FUNC func,gpointer arg)
{
	VWM	*vwm;

	vwm=vwm_get_instance();

	vwm->hook_enter=func;
	vwm->hook_enter_arg=arg;

	return;
}

void vwm_hook_leave(VIPER_FUNC func,gpointer arg)
{
	VWM	*vwm;

	vwm=vwm_get_instance();

	vwm->hook_leave=func;
	vwm->hook_leave_arg=arg;

	return;
}
